from decimal import Decimal as D
import pytest

from eth_utils import from_wei, to_checksum_address

from moonwalking import wallets
from moonwalking.blocks.exc import NotEnoughAmountError
from moonwalking.main import Ethereum
from moonwalking.testing import ETH_MAIN_ADDR, send_eth


def test_create_addr():
    addr, pk = wallets.create_addr('eth')
    assert addr
    assert pk


async def test_not_enough_amount():
    eth = Ethereum()
    addr1, priv1 = await wallets.create_wallet('eth')
    addr2, priv2 = await eth.create_wallet()

    assert await eth.get_balance(addr1) == D(0)
    assert await send_eth(addr1, D(1))
    assert await eth.get_balance(addr1) == D(1)

    with pytest.raises(NotEnoughAmountError):
        await eth.send_money(priv1, [(addr2, D(2))])


async def test_send_money(fee_mocker):
    eth = Ethereum()
    addr1, priv1 = await eth.create_wallet()
    addr2, priv2 = await eth.create_wallet()

    assert await eth.get_balance(addr1) == 0
    assert await send_eth(addr1, D(1))
    assert await eth.get_balance(addr1) == D(1)
    assert await eth.get_balance(addr2) == 0
    assert await eth.send_money(priv1, [(addr2, D('0.5'))])
    fee = D('0.00021')
    assert await eth.get_balance(addr1) == D('0.5')
    assert await eth.get_balance(addr2) == D('0.5') - fee


async def test_send_money_to_multiple_recipients(fee_mocker):
    eth = Ethereum()
    fee = D('0.00021')
    addr1, priv1 = await eth.create_wallet()
    addr2, priv2 = await eth.create_wallet()
    addr3, priv3 = await eth.create_wallet()
    addr4, priv4 = await eth.create_wallet()

    assert await eth.get_balance(addr1) == 0
    assert await send_eth(addr1, D(1))
    assert await eth.get_balance(addr1) == D(1)
    assert await eth.get_balance(addr2) == 0
    assert await eth.get_balance(addr3) == 0
    assert await eth.send_money(priv1, [(addr2, D('0.1')), (addr3, D('0.2'))])
    assert await eth.get_balance(addr1) == D('0.7')
    assert await eth.get_balance(addr2) == D('0.1') - fee
    assert await eth.get_balance(addr3) == D('0.2') - fee
    assert await eth.send_money(priv1, [(addr4, D('0.7'))])
    assert await eth.get_balance(addr1) == 0
    assert await eth.get_balance(addr4) == D('0.7') - fee


async def test_send_money_contract(fee_mocker):
    eth = Ethereum()
    addr1, priv1 = await eth.create_wallet()

    assert await send_eth(addr1, D(10))
    assert await eth.get_balance(addr1) == D(10)

    # deploy a simple contract that accepts eth
    tx_hash = await eth.post('eth_sendTransaction', {
        'from': ETH_MAIN_ADDR,
        'gas': 4000000,
        'gasPrice': 20,
        'data': (
            '0x608060405234801561001057600080fd5b50336000806101000'
            'a81548173ffffffffffffffffffffffffffffffffffffffff02191'
            '6908373ffffffffffffffffffffffffffffffffffffffff1602179'
            '05550610292806100606000396000f300608060405260043610610'
            '04c576000357c01000000000000000000000000000000000000000'
            '00000000000000000900463ffffffff1680633ccfd60b1461009c5'
            '780638da5cb5b146100b3575b3373fffffffffffffffffffffffff'
            'fffffffffffffff167fd4f43975feb89f48dd30cabbb32011045be'
            '187d1e11c8ea9faa43efc352825193460405180828152602001915'
            '05060405180910390a2005b3480156100a857600080fd5b506100b'
            '161010a565b005b3480156100bf57600080fd5b506100c86102415'
            '65b604051808273fffffffffffffffffffffffffffffffffffffff'
            'f1673ffffffffffffffffffffffffffffffffffffffff168152602'
            '00191505060405180910390f35b6000809054906101000a900473f'
            'fffffffffffffffffffffffffffffffffffffff1673fffffffffff'
            'fffffffffffffffffffffffffffff163373fffffffffffffffffff'
            'fffffffffffffffffffff1614151561016557600080fd5b6000809'
            '054906101000a900473fffffffffffffffffffffffffffffffffff'
            'fffff1673ffffffffffffffffffffffffffffffffffffffff16337'
            '3ffffffffffffffffffffffffffffffffffffffff161415156101c'
            '057600080fd5b6000809054906101000a900473fffffffffffffff'
            'fffffffffffffffffffffffff1673fffffffffffffffffffffffff'
            'fffffffffffffff166108fc3073fffffffffffffffffffffffffff'
            'fffffffffffff16319081150290604051600060405180830381858'
            '888f1935050505015801561023e573d6000803e3d6000fd5b50565'
            'b6000809054906101000a900473fffffffffffffffffffffffffff'
            'fffffffffffff16815600a165627a7a7230582085608321fc053d1'
            '45b06de95c945cafcb58eceb5fbd45a892860ac0c5ceae0c50029'
        )
    })
    receipt = await eth.post(
        'eth_getTransactionReceipt',
        tx_hash
    )
    addr2 = to_checksum_address(receipt['contractAddress'])
    assert await eth.send_money(priv1, [(addr2, D('0.5'))])
    fee = D(50000) * D(from_wei(10, 'gwei'))
    assert await eth.get_balance(addr2) == D('0.5') - fee



from abc import ABC, abstractmethod
from decimal import Decimal as D
from typing import Dict, Tuple, List


class BlockBaseError(Exception):
    error = None

    def __init__(self, data=None):
        self.data = data
        super().__init__()


class EthereumError(BlockBaseError):
    error = 'unknown_error'


class ReplacementTransactionError(Exception):
    pass


class NotEnoughAmountError(BlockBaseError):
    error = 'not_enough_amount'


class BaseBlock(ABC):
    CCY: str = None
    BLOCKS: Dict[str, 'BaseBlock'] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.register()

    @classmethod
    def register(cls):
        if cls.CCY in BaseBlock.BLOCKS:
            raise ValueError(f"Block already there for {cls.CCY}")
        if cls.CCY:
            BaseBlock.BLOCKS[cls.CCY] = cls()

    @abstractmethod
    def validate_addr(self, addr: str):
        pass

    @abstractmethod
    async def create_addr(self) -> Tuple[str, str]:
        pass

    @abstractmethod
    async def create_wallet(self) -> Tuple[str, str]:
        pass

    async def send_money(self, priv: str, addrs: List[Tuple[str, D]]) -> str:
        tx = await self.build_tx(priv, addrs)
        signed = self.sign_tx(priv, tx)
        return await self.broadcast_tx(signed)

    @abstractmethod
    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        pass

    @abstractmethod
    def sign_tx(self, priv, tx):
        pass

    @abstractmethod
    async def broadcast_tx(self, tx):
        pass

    @abstractmethod
    async def get_balance(self, addr: str) -> D:
        pass



from eth_utils.address import is_address

from cashaddress.convert import is_valid

from pycoin.key.validate import is_address_valid


class CcyHelper:
    def __init__(self, ccy, use_testnet=False):
        self.ccy = ccy
        self.network = 'testnet' if use_testnet else 'mainnet'

    def validate_addr(self, addr):
        try:
            return getattr(self, f'validate_{self.ccy.lower()}')(addr)
        except AttributeError:
            return False

    def validate_btc(self, addr):
        if self.network == 'testnet':
            return is_address_valid(addr) == 'XTN'
        return is_address_valid(addr) == 'BTC'

    def validate_ltc(self, addr):
        if self.network == 'testnet':
            return is_address_valid(addr) == 'XTN'
        return is_address_valid(addr) == 'LTC'

    @classmethod
    def validate_eth(cls, addr):
        return is_address(addr)

    @classmethod
    def validate_bch(cls, addr):
        return is_valid(addr)

    @classmethod
    def validate_lnd(cls, addr):
        return is_address(addr)



import asyncio

import pytest
from eth_utils.address import to_checksum_address
from eth_utils.currency import to_wei

from moonwalking.testing import create_lnd_contract


@pytest.fixture(autouse=True)
def loop():
    """Return an instance of the event loop."""
    loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture()
async def lnd_helper(mocker):
    contract_addr = await create_lnd_contract()
    mocker.patch(
        'moonwalking.blocks.eth_generic.EthereumGeneric.get_contract_addr',
        lambda self: to_checksum_address(contract_addr),
    )
    yield lnd_helper


async def calc_fee_mock(self, tx):
    return 500


async def get_gas_price_mock(self):
    return to_wei(10, 'gwei')


@pytest.fixture()
async def fee_mocker(mocker):
    mocker.patch(
        'moonwalking.main.Bitcoin.calc_fee',
        calc_fee_mock
    )
    mocker.patch(
        'moonwalking.main.Litecoin.calc_fee',
        calc_fee_mock
    )
    mocker.patch(
        'moonwalking.main.BitcoinCash.calc_fee',
        lambda x, y, z: 500
    )
    mocker.patch(
        'moonwalking.blocks.eth_generic.EthereumGeneric.get_gas_price',
        get_gas_price_mock
    )



from .main import BLOCKS


def block(currency):
    block = BLOCKS.get(currency)
    if not block:
        raise ValueError(f'Blockchain {currency} not supported')
    return block


def validate_addr(currency, addr):
    return block(currency).validate_addr(addr)


def create_addr(currency):
    return block(currency).create_addr()


async def create_wallet(currency):
    return await block(currency).create_wallet()



import logging
from decimal import Decimal as D
from typing import List, Tuple

from bitcash import PrivateKeyTestnet, PrivateKey
from bitcash.network.meta import Unspent
from bitcash.transaction import estimate_tx_fee, create_p2pkh_transaction
from bitcoin.core import COIN

from cashaddress.convert import is_valid, to_legacy_address

from eth_account import Account
from eth_utils import from_wei

from pycoin.key.validate import is_address_valid
from pycoin.tx import Tx

from . import settings
from .blocks.base import BaseBlock
from .blocks.bitcoin_generic import BitcoinGeneric, to_string
from .blocks.eth_generic import EthereumGeneric
from .blocks.exc import NotEnoughAmountError
from .blocks.fee import FeeStation
from .utils import GeneralError, rand_str

logger = logging.getLogger(__name__)

DECIMALS = pow(10, 18)  # Todo: Rename me pls.
BLOCKS = BaseBlock.BLOCKS


class BlockError(Exception):
    pass


class WrongKeyError(BlockError):
    pass


class WrongAddressError(BlockError):
    pass


class InsufficientFoundsError(BlockError):
    pass


class Dummycoin(BaseBlock):
    """used in tests"""
    ADDRESSES = {}
    PRIV_KEYS = {}

    def validate_addr(self, addr):
        if addr in self.ADDRESSES:
            return addr

    def create_addr(self):
        addr = rand_str()
        priv_key = f'1_{addr}'
        self.ADDRESSES[addr] = D(0)
        self.PRIV_KEYS[priv_key] = addr
        return addr, priv_key

    async def create_wallet(self):
        return self.create_addr()

    async def get_balance(self, addr):
        if addr not in self.ADDRESSES:
            raise WrongAddressError(addr)
        return self.ADDRESSES[addr]

    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        for addr, amount in addrs:
            if not self.validate_addr(addr):
                raise WrongAddressError(addr)

        wallet_addr = priv[2:]
        balance = self.ADDRESSES[wallet_addr]

        sum_amounts = sum(addr[1] for addr in addrs)

        if balance < sum_amounts:
            raise InsufficientFoundsError(str(balance))

        self.ADDRESSES[wallet_addr] -= sum_amounts
        for addr, amount in addrs:
            self.ADDRESSES[addr] += amount

        return rand_str()

    def sign_tx(self, priv, tx):
        return tx

    async def broadcast_tx(self, tx):
        return tx


class Bitcoin(BitcoinGeneric):
    CCY = 'btc'
    MAX_FEE = 100
    URL = settings.BITCOIN_URL
    NET_WALLET = 'btctest' if settings.USE_TESTNET else 'btc'
    NETWORK = 'testnet' if settings.USE_TESTNET else 'mainnet'
    NETCODE = 'XTN' if settings.USE_TESTNET else 'BTC'

    def validate_addr(self, addr):
        if settings.USE_TESTNET:
            if is_address_valid(addr) == 'XTN':
                return addr
        if is_address_valid(addr) == 'BTC':
            return addr

    async def calc_fee(self, tx: Tx) -> int:
        if not settings.BITCOIN_FEE:
            fee_station = FeeStation('btc')
            transaction_fee = await fee_station.get_fee()
            transaction_fee = min(self.MAX_FEE, transaction_fee)
        else:
            transaction_fee = int(settings.BITCOIN_FEE)
        tx_size = self.calculate_tx_size(tx)
        return transaction_fee * tx_size


class Litecoin(BitcoinGeneric):
    CCY = 'ltc'
    URL = settings.LITECOIN_URL
    NET_WALLET = 'ltctest' if settings.USE_TESTNET else 'ltc'
    NETWORK = 'testnet' if settings.USE_TESTNET else 'mainnet'
    NETCODE = 'XTN' if settings.USE_TESTNET else 'LTC'

    def validate_addr(self, addr):
        if settings.USE_TESTNET:
            if is_address_valid(addr) == 'XTN':
                return addr
        if is_address_valid(addr) == 'LTC':
            return addr

    async def calc_fee(self, tx: Tx) -> int:
        tx_size = self.calculate_tx_size(tx)
        return int(settings.LITECOIN_FEE) * tx_size


class BitcoinCash(BitcoinGeneric):
    CCY = 'bch'
    URL = settings.BITCOIN_CASH_URL
    NET_WALLET = 'btctest' if settings.USE_TESTNET else 'btc'
    NETWORK = 'testnet' if settings.USE_TESTNET else 'mainnet'
    KEY_CLASS = PrivateKeyTestnet if settings.USE_TESTNET else PrivateKey

    def validate_addr(self, addr):
        if is_valid(addr):
            return self.to_legacy_address(addr)

    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        # Todo: Need to break this up.
        key = self.KEY_CLASS(priv)
        addr_from = self.to_legacy_address(key.address)
        unspent_obj_list = await self._get_obj_unspent_list(addr_from)

        payables = [
            (self.to_legacy_address(addr), amount * COIN)
            for addr, amount in addrs
        ]
        total_out = sum(amount for addr, amount in payables)
        total_unspent = sum(D(unspent.amount) for unspent in unspent_obj_list)
        remaining = total_unspent - total_out

        if total_out > total_unspent:
            raise NotEnoughAmountError()

        fee = self.calc_fee(len(unspent_obj_list), len(addrs))
        fee_per_tx_out, extra_count = divmod(fee, len(addrs))

        calc_addrs = []
        for addr, amount in payables:
            amount -= fee_per_tx_out
            if extra_count > 0:
                amount -= 1
            if amount < 1:
                raise NotEnoughAmountError()
            calc_addrs.append((addr, int(amount)))
        remaining = int(remaining)
        if remaining > 0:
            calc_addrs.append((addr_from, remaining))

        return create_p2pkh_transaction(key, unspent_obj_list, calc_addrs)

    def sign_tx(self, priv, tx):
        return tx

    async def broadcast_tx(self, tx):
        return await self.post('sendrawtransaction', tx)

    async def get_balance(self, addr):
        unspent_list = await self._get_raw_unspent_list(addr)
        d = D(str(sum(
            D(str(unspent['amount'])) for unspent in unspent_list
        )))
        return self.normalize_decimal(d)

    def create_addr(self):
        key = self.KEY_CLASS()
        return self.to_legacy_address(key.address), to_string(key.to_wif())

    @staticmethod
    def calc_fee(n_in, n_out):
        return estimate_tx_fee(
            n_in,
            n_out,
            int(settings.BITCOIN_CASH_FEE),
            False,
        )

    async def _get_raw_unspent_list(self, addr, confirmations=1):
        addr = self.to_legacy_address(addr)
        list_unspent = await self.post(
            'listunspent',
            confirmations,
            9999999,
            [addr],
        )
        return list_unspent

    async def _get_obj_unspent_list(self, addr):
        unspent_list = await self._get_raw_unspent_list(addr)
        return [
            Unspent(
                int(D(str(unspent['amount'])) * COIN),
                unspent['confirmations'],
                unspent['scriptPubKey'],
                unspent['txid'],
                unspent['vout']
            )
            for unspent in unspent_list
        ]

    def to_legacy_address(self, addr):
        return to_string(to_legacy_address(addr))

    @staticmethod
    def normalize_decimal(d):
        return d.to_integral() if d == d.to_integral() else d.normalize()


class Ethereum(EthereumGeneric):
    CCY = 'eth'

    async def get_balance(self, addr):
        return await self.get_eth_balance(addr)

    async def create_wallet(self):
        return self.create_addr()


class Lendingblock(EthereumGeneric):
    CCY = 'lnd'
    LND_WALLETS_TOPUP_TRANS_NO = int(settings.LND_WALLETS_TOPUP_TRANS_NO)

    async def get_balance(self, addr):
        method_hash = self.get_method_hash('balanceOf')
        addr_hash = self.get_addr_hash(addr)
        result = await self.post('eth_call', {
            'data': method_hash + addr_hash,
            'to': self.get_contract_addr(),
        }, 'latest')
        return D(int(result, 16) / DECIMALS)

    async def create_wallet(self):
        addr, priv = self.create_addr()
        price = await self.get_gas_price()
        single_tx_price = price * self.MAX_GAS
        single_tx_price = from_wei(single_tx_price, 'ether')
        try:
            await self.send_eth(
                settings.BUFFER_ETH_PRIV,
                [(addr, single_tx_price * self.LND_WALLETS_TOPUP_TRANS_NO)],
            )
        except NotEnoughAmountError:
            raise GeneralError("not enough eth in buffer wallet")

        return addr, priv

    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        # Todo: Deduplicate with EthereumGeneric.build_tx.
        addr_from = Account.privateKeyToAccount(priv).address
        nonce = await self.get_transaction_count(addr_from)
        contract_addr = self.get_contract_addr()
        return [
            (await self.get_transaction_dict(
                priv,
                contract_addr,
                0,
                nonce + i,
                self.make_lnd_transfer_data(addr, amount),
                False
            ))
            for i, (addr, amount) in enumerate(addrs)
        ]



import pycoin

from moonwalking.main import Litecoin


async def test_litecoincreate_wallet():
    ltc = Litecoin()
    ltc.NET_WALLET = 'ltc'
    addr1, priv1 = await ltc.create_wallet()

    assert pycoin.encoding.is_hashed_base58_valid(addr1) is True
    assert pycoin.encoding.is_hashed_base58_valid(priv1) is True
    assert pycoin.key.validate.is_wif_valid(priv1) == 'LTC'
    assert pycoin.key.validate.is_private_bip32_valid(priv1) is None
    assert pycoin.key.validate.is_address_valid(addr1) == 'LTC'
    assert pycoin.key.Key.from_text(priv1).address() == addr1
    assert addr1[0] == 'L'
    assert priv1[0] == 'T'

    priv2 = 'TB648cs5JzYD2QjCbGdTJ4X3BM1uBPrXFjnJq1UNwZ9VTwWvKanv'
    addr2 = 'LcNs562P7ZEtAhRcQTabwNEa9QETvEiLzz'
    assert pycoin.key.Key.from_text(priv2).address() == addr2


async def test_litecoin_validate_addr(mocker):
    ltc = Litecoin()
    mocker.patch('moonwalking.settings.USE_TESTNET', None)

    # testnet address
    assert ltc.validate_addr('mn8eCaT46d8mEn62ussMtE467J4mSgu5zA') is None
    # P2PKH bitcoin address starting with 1
    assert ltc.validate_addr('1BTm2i94LVDs6vMkf4qDs1Ato1NszQiE4z') is None
    # P2SH address
    assert ltc.validate_addr('3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLy') is None
    # Bech32 address
    assert ltc.validate_addr(
        'bc1qar0srrr7xfkvy5l643lydnw9re59gtzzwf5mdq'
        ) is None
    # litecoin address
    assert ltc.validate_addr('LcNs562P7ZEtAhRcQTabwNEa9QETvEiLzz') == \
        'LcNs562P7ZEtAhRcQTabwNEa9QETvEiLzz'
    # all lowercase
    assert ltc.validate_addr('lcns562p7zetahrcqtabwnea9qetveilzz') is None
    # all uppercase
    assert ltc.validate_addr('LCNS562P7ZETAHRCQTABWNEA9QETVEILZZ') is None
    # wrong checksum
    assert ltc.validate_addr('LcNs562P7ZEtAhRcQTabwNEa9QETvEiLza') is None
    # litecoin P2SH address
    assert ltc.validate_addr('MV5rN5EcX1imDS2gEh5jPJXeiW5QN8YrK3') is None



from moonwalking.blocks.ccyhelper import CcyHelper


def test_ccyhelper():
    ccyhelper = CcyHelper('btc', use_testnet=True)
    assert ccyhelper.validate_addr('mtXWDB6k5yC5v7TcwKZHB89SUp85yCKshy')

    ccyhelper = CcyHelper('btc')
    assert not ccyhelper.validate_addr('mtXWDB6k5yC5v7TcwKZHB89SUp85yCKshy')

    ccyhelper = CcyHelper('btc')
    assert ccyhelper.validate_addr('1F1tAaz5x1HUXrCNLbtMDqcw6o5GNn4xqX')

    ccyhelper = CcyHelper('ltc')
    assert ccyhelper.validate_addr('LWSygPfS6FEiDqdj2xmVF8CSZJREo4LbKd')

    ccyhelper = CcyHelper('ltc', use_testnet=True)
    assert ccyhelper.validate_addr('mgTbDyNGwJeewjdXmU9cRQe8WDauVqn4WK')

    ccyhelper = CcyHelper('bch')
    assert ccyhelper.validate_addr('mhstAGNEZYNxwpwgAqwX31sK2TK7SntHCK')

    ccyhelper = CcyHelper('eth')
    assert ccyhelper.validate_addr(
        '0x0dE0BCb0703ff8F1aEb8C892eDbE692683bD8030'
    )

    ccyhelper = CcyHelper('lnd')
    assert ccyhelper.validate_addr(
        '0x0dE0BCb0703ff8F1aEb8C892eDbE692683bD8030'
    )

    ccyhelper = CcyHelper('fake_currency')
    assert not ccyhelper.validate_addr('1F1tAaz5x1HUXrCNLbtMDqcw6o5GNn4xqX')



import re

import eth_utils
from eth_account import Account

from moonwalking.main import Lendingblock
from moonwalking import wallets


async def test_lnd_create_wallet(fee_mocker, lnd_helper):
    lnd = Lendingblock()
    addr1, priv1 = await lnd.create_wallet()

    pattern_address = re.compile('^(0x)[0-9a-fA-F]{40}$')
    pattern_private = re.compile('^(0x)[0-9a-fA-F]{64}$')
    assert pattern_address.match(addr1)
    assert pattern_private.match(priv1)
    assert eth_utils.is_checksum_address(addr1) is True
    assert Account.privateKeyToAccount(priv1).address == addr1

    priv2 = (
        '0xb25c7db31feed9122727bf0939dc769a96564b2de4c4726d035b36ecf1e5b364')
    addr2 = '0x5ce9454909639D2D17A3F753ce7d93fa0b9aB12E'
    assert Account.privateKeyToAccount(priv2).address == addr2


async def test_lnd_validate_addr():
    lnd = Lendingblock()

    # do not pad missing length
    assert lnd.validate_addr('0x0') is None
    # all uppercase valid checksummed
    assert wallets.validate_addr(
        'lnd',
        '0x52908400098527886E0F7030069857D2E4169EE7'
        ) == '0x52908400098527886E0F7030069857D2E4169EE7'
    # all lowercase valid checksummed
    assert lnd.validate_addr(
        '0xde709f2102306220921060314715629080e2fb77'
        ) == '0xde709f2102306220921060314715629080e2fb77'
    # normal checksummed address
    assert lnd.validate_addr(
        '0xd3CdA913deB6f67967B99D67aCDFa1712C293601'
        ) == '0xd3CdA913deB6f67967B99D67aCDFa1712C293601'
    # wrong checksummed address
    assert lnd.validate_addr(
        '0xd3cdA913deB6f67967B99D67aCDFa1712C293601'
        ) is None
    # all lowercase
    assert lnd.validate_addr(
        '0xd3cda913deb6f67967b99d67acdfa1712c293601'
        ) == '0xd3cda913deb6f67967b99d67acdfa1712c293601'
    # all uppercase
    assert lnd.validate_addr(
        '0xD3CDA913DEB6F67967B99D67ACDFA1712C293601'
        ) == '0xD3CDA913DEB6F67967B99D67ACDFA1712C293601'
    # starts with 00
    assert lnd.validate_addr(
        '0x00908400098527886E0F7030069857d2e4169eE7'
        ) == '0x00908400098527886E0F7030069857d2e4169eE7'
    # do not pad missing length
    assert lnd.validate_addr(
        '0x908400098527886E0F7030069857d2e4169eE7'
        ) is None
    # ends with 00
    assert lnd.validate_addr(
        '0x52908400098527886e0F7030069857d2E4169e00'
        ) == '0x52908400098527886e0F7030069857d2E4169e00'
    # do not pad missing length
    assert lnd.validate_addr(
        '0x52908400098527886e0F7030069857d2E4169e'
        ) is None



from decimal import Decimal as D

from bitcoin.core import COIN

from moonwalking.main import Litecoin


async def test_basic_ltc_ops(fee_mocker):
    ltc = Litecoin()
    addr1, priv1 = await ltc.create_wallet()
    addr2, priv2 = await ltc.create_wallet()
    addr3, priv3 = await ltc.create_wallet()

    assert ltc.validate_addr(addr1)
    assert ltc.validate_addr(addr2)
    assert ltc.validate_addr(addr3)

    await ltc.post('sendtoaddress', addr1, 100000 / COIN)
    await ltc.post('generate', 1)
    assert (await ltc.get_balance(addr1)) == D(100000) / COIN

    tx_id = await ltc.send_money(
        priv1,
        [(addr2, D(5000) / COIN), (addr3, D(15000) / COIN)]
    )
    assert tx_id
    await ltc.post('generate', 1)

    # 500 fee splited equally onto 2 recievers:
    assert (await ltc.get_balance(addr1)) == D(100000 - 20000) / COIN
    assert (await ltc.get_balance(addr2)) == D(5000 - 250) / COIN
    assert (await ltc.get_balance(addr3)) == D(15000 - 250) / COIN



import random
import string


def rand_str(n=10):
    return ''.join(random.choices(
        string.ascii_lowercase + string.digits,
        k=n,
    ))


class GeneralError(Exception):
    """raised when something is wrong and we cannot handle it properly"""



import io
from typing import List, Tuple
from binascii import unhexlify
from decimal import Decimal as D

from aiohttp import ClientSession

from bitcoin import SelectParams
from bitcoin.core import COIN, lx, COutPoint
from bitcoin.core.script import CScript
from bitcoin.wallet import CBitcoinAddress

from pycoin.key import Key
from pycoin.tx.Tx import Tx
from pycoin.tx.Spendable import Spendable
from pycoin.tx.TxOut import TxOut
from pycoin.tx.tx_utils import sign_tx
from pycoin.ui import standard_tx_out_script

from pywallet.wallet import create_wallet

from .base import BaseBlock
from .exc import NotEnoughAmountError


def to_string(v):
    return v.decode('utf-8') if isinstance(v, bytes) else v


class BitcoinGeneric(BaseBlock):
    NETWORK = None
    URL = None
    NETCODE = None
    NET_WALLET = None

    def __init__(self):
        SelectParams(self.NETWORK)

    def get_data(self, method, *params):
        return {
            'version': '1.1',
            'method': method,
            'params': list(params),
            'id': self.NETWORK
        }

    async def post(self, *args):
        async with ClientSession() as sess:
            async with sess.post(self.URL, json=self.get_data(*args)) as res:
                resp_dict = await res.json()
                return resp_dict['result']

    def create_addr(self):
        wallet = create_wallet(self.NET_WALLET)
        return to_string(wallet['address']), to_string(wallet['wif'])

    async def create_wallet(self):
        addr, pk = self.create_addr()
        await self.post('importaddress', addr, '', False)
        return addr, pk

    async def get_listunspent_for_addr(self, addr, confirmations=1):
        res = await self.post('listunspent', confirmations, 9999999, [addr])
        for unspent in res:
            coutpoint = COutPoint(lx(unspent['txid']), unspent['vout'])
            cscript = CScript(unhexlify(unspent['scriptPubKey']))
            unspent['outpoint'] = coutpoint
            unspent['address'] = CBitcoinAddress(unspent['address'])
            unspent['scriptPubKey'] = cscript
            unspent['amount'] = int(unspent['amount'] * COIN)
        return res

    async def get_spendable_list_for_addr(self, addr: str) -> List[Spendable]:
        unspent_list = await self.get_listunspent_for_addr(addr)
        return [
            Spendable(
                unspent['amount'],
                bytes(unspent['scriptPubKey']),
                unspent['outpoint'].hash,
                unspent['outpoint'].n,
            )
            for unspent in unspent_list
        ]

    async def get_balance(self, addr):
        unspent_list = await self.get_listunspent_for_addr(addr)
        return D(sum(unspent['amount'] for unspent in unspent_list)) / COIN

    def calc_fee(self, tx: Tx) -> D:
        raise NotImplementedError

    @staticmethod
    def calculate_tx_size(tx: Tx) -> int:
        s = io.BytesIO()
        tx.stream(s)
        return len(s.getvalue())

    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        """
        We distribute fee equally on every recipient by reducing the amount
        of money they will receive. The rest will go back to the sender
        as a change.

        :param priv: WIF private key of sender -> str
        :param addrs: distribution -> [(addr1, amount1), (addr2, amount2),...]
        :return: transaction id -> str
        """
        addr = Key.from_text(priv).address()

        spendables = await self.get_spendable_list_for_addr(addr)
        addrs.append((addr, D(0)))

        txs_out = []
        for payable in addrs:
            bitcoin_address, coin_value = payable
            coin_value *= COIN
            script = standard_tx_out_script(bitcoin_address)
            txs_out.append(TxOut(coin_value, script))

        txs_in = [spendable.tx_in() for spendable in spendables]
        tx = Tx(version=1, txs_in=txs_in, txs_out=txs_out, lock_time=0)

        tx.set_unspents(spendables)

        fee = await self.calc_fee(tx)

        fee_per_tx_out, extra_count = divmod(fee, len(tx.txs_out) - 1)

        total_coin_value = sum(
            spendable.coin_value
            for spendable in tx.unspents
        )
        coins_allocated = sum(tx_out.coin_value for tx_out in tx.txs_out)

        if coins_allocated > total_coin_value:
            raise NotEnoughAmountError()

        for tx_out in tx.txs_out:
            if tx_out.address(netcode=self.NETCODE) == addr:
                tx_out.coin_value = total_coin_value - coins_allocated
            else:
                tx_out.coin_value -= fee_per_tx_out
                if extra_count > 0:
                    tx_out.coin_value -= 1
                    extra_count -= 1
                if tx_out.coin_value < 1:
                    raise NotEnoughAmountError()
        return tx

    def sign_tx(self, priv, tx):
        sign_tx(tx, wifs=[priv])
        return tx

    async def broadcast_tx(self, tx):
        return await self.post('sendrawtransaction', tx.as_hex())



class BlockBaseError(Exception):
    def __init__(self, data=None):
        self.data = data
        super().__init__()

    error = None


class EthereumError(BlockBaseError):
    error = 'unknown_error'


class ReplacementTransactionError(Exception):
    pass


class NotEnoughAmountError(BlockBaseError):
    error = 'not_enough_amount'



from datetime import datetime, timedelta

from aiohttp import ClientSession

from eth_utils.currency import to_wei

from .. import settings


class FeeStation:

    FEE_CACHES = {}
    BITCOIN_FEE_URL = 'https://bitcoinfees.earn.com/api/v1/fees/recommended'
    ETH_GAS_PRICE_URL = 'https://ethgasstation.info/json/ethgasAPI.json'

    def __init__(self, currency):
        self.currency = currency.lower()

    async def get_fee(self):
        fee, dt = self.FEE_CACHES.get(self.currency, (None, None))
        valid_till = datetime.utcnow() - timedelta(
            minutes=int(settings.FEE_CACHE_TIME)
        )
        if not fee or dt < valid_till:
            method = f'get_{self.currency}_fee'
            value = await getattr(self, method)()
            self.FEE_CACHES[self.currency] = (value, datetime.utcnow())
            return value
        return fee

    @staticmethod
    async def get(url):
        async with ClientSession() as session:
            async with session.get(url) as resp:
                return await resp.json()

    async def get_btc_fee(self):
        resp_dict = await self.get(self.BITCOIN_FEE_URL)
        fatest_fee = int(resp_dict['fastestFee'])
        return fatest_fee

    async def get_eth_fee(self):
        resp_dict = await self.get(self.ETH_GAS_PRICE_URL)
        average = int(resp_dict['average'] / 10)
        return to_wei(average, 'gwei')



from eth_keys.datatypes import PrivateKey, PublicKey
from eth_utils import to_wei

from moonwalking.main import Ethereum
from . import settings


def private_key_to_checksum_address(key):
    if key.startswith('0x'):
        key = key[2:]
    return PublicKey.from_private(
        PrivateKey(bytes.fromhex(key))
    ).to_checksum_address()


ETH_MAIN_ADDR = private_key_to_checksum_address(settings.BUFFER_ETH_PRIV)
eth = Ethereum()


async def send_eth(addr, amount):
    nonce = await eth.post(
        'eth_getTransactionCount',
        ETH_MAIN_ADDR,
    )
    tx = {
        'from': ETH_MAIN_ADDR,
        'to': addr,
        'value': to_wei(amount, 'ether'),
        'gas': 22000,
        'gasPrice': to_wei(8, 'gwei'),
        'chainId': 1,
        'nonce': nonce,
    }
    return await eth.post('eth_sendTransaction', tx)


async def create_lnd_contract():
    tx_hash = await eth.post('eth_sendTransaction', {
        'from': ETH_MAIN_ADDR,
        'gas': 4000000,
        'gasPrice': 100,
        'data': settings.LND_CONTRACT['bytecode'],
    })
    receipt = await eth.post(
        'eth_getTransactionReceipt',
        tx_hash
    )
    return receipt['contractAddress']



from decimal import Decimal as D

from eth_account import Account

from moonwalking.main import Lendingblock
from moonwalking.testing import send_eth

PRIV_KEY = '0x869844d42d74171d1c5e71ecd8964118e68f610af94047fd1e98afb4df1c5e1b'


async def test_lnd(lnd_helper, fee_mocker):
    lnd = Lendingblock()
    account = Account.privateKeyToAccount(PRIV_KEY).address

    # initial balance is 1 billion which is pow(10, 9)
    decimals = await lnd.call_contract_method('decimals', to_int=True)
    assert decimals == 18
    initial_balance = await lnd.get_balance(account)
    total_supply = await lnd.call_contract_method(
        'totalSupply',
        to_int=True,
    )
    assert initial_balance == D(total_supply / pow(10, decimals))
    assert initial_balance == D(pow(10, 9))

    name = await lnd.call_contract_method(
        'name',
        to_string=True,
    )
    assert name == 'Lendingblock'

    symbol = await lnd.call_contract_method(
        'symbol',
        to_string=True,
    )
    assert symbol == 'LND'

    addr1, priv1 = await lnd.create_wallet()
    addr2, priv2 = await lnd.create_wallet()
    addr3, priv3 = await lnd.create_wallet()

    assert lnd.validate_addr(addr1)
    assert lnd.validate_addr(addr2)
    assert lnd.validate_addr(addr3)

    # we need some ETH to pay transaction fees
    assert await send_eth(addr1, D(1))

    tx_ids = await lnd.send_money(
        PRIV_KEY,
        [(addr1, D(1000)), (addr2, D(2000))],
    )

    assert len(tx_ids) == 2

    # new_balance is 1 billion - 1000 - 2000
    assert await lnd.get_balance(account) == D(1000000000 - 1000 - 2000)
    assert await lnd.get_balance(addr1) == D(1000)
    assert await lnd.get_balance(addr2) == D(2000)
    assert await lnd.get_balance(addr3) == D(0)

    assert await lnd.send_money(priv1, [(addr3, D(400))])
    assert await lnd.get_balance(addr1) == D(1000 - 400)
    assert await lnd.get_balance(addr3) == D(400)


async def test_get_addr_hash():
    addr = '0x4092678e4E78230F46A1534C0fbc8fA39780892B'
    lnd = Lendingblock()
    assert lnd.get_addr_hash(addr) == \
        '0000000000000000000000004092678e4e78230f46a1534c0fbc8fa39780892b'
    assert lnd.get_addr_hash(addr[2:]) == ''



import pycoin

from moonwalking.main import Bitcoin


async def test_bitcoin_create_wallet():
    bitcoin = Bitcoin()
    bitcoin.NET_WALLET = 'btc'
    addr1, priv1 = await bitcoin.create_wallet()

    assert pycoin.encoding.is_hashed_base58_valid(addr1) is True
    assert pycoin.encoding.is_hashed_base58_valid(priv1) is True
    assert pycoin.encoding.is_valid_bitcoin_address(addr1) is True
    assert pycoin.encoding.is_valid_wif(priv1) is True
    assert pycoin.key.validate.is_wif_valid(priv1) == 'BTC'
    assert pycoin.key.validate.is_private_bip32_valid(priv1) is None
    assert pycoin.key.validate.is_address_valid(addr1) == 'BTC'
    assert pycoin.key.Key.from_text(priv1).address() == addr1
    assert addr1[0] == '1'
    assert priv1[0] in ['L', 'K']

    priv2 = 'L3XJp5UxLouN3Mw6ycJ4fJT678gjhERx9yfTfzYpVNqom1g7i5Nu'
    addr2 = '1BTm2i94LVDs6vMkf4qDs1Ato1NszQiE4z'
    assert pycoin.key.Key.from_text(priv2).address() == addr2


async def test_bitcoin_validate_addr(mocker):
    bitcoin = Bitcoin()
    mocker.patch('moonwalking.settings.USE_TESTNET', None)

    # testnet address
    assert bitcoin.validate_addr('mn8eCaT46d8mEn62ussMtE467J4mSgu5zA') is None
    # P2PKH address starting with 1
    assert bitcoin.validate_addr('1BTm2i94LVDs6vMkf4qDs1Ato1NszQiE4z') == \
        '1BTm2i94LVDs6vMkf4qDs1Ato1NszQiE4z'
    # all lowercase
    assert bitcoin.validate_addr('1btm2i94lvds6vmkf4qds1ato1nszqie4z') is None
    # all uppercase
    assert bitcoin.validate_addr('1BTM2I94LVDS6VMKF4QDS1ATO1NSZQIE4Z') is None
    # wrong checksum
    assert bitcoin.validate_addr('1BTm2i94LVDs6vMkf4qDs1Ato1NszQiE4a') is None
    # P2SH address
    assert bitcoin.validate_addr('3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLy') == \
        '3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLy'
    # wrong checksum
    assert bitcoin.validate_addr('3J98t1WpEZ73CNmQviecrnyiWrnqRhWNLa') is None
    # Bech32 address
    assert bitcoin.validate_addr(
        'bc1qar0srrr7xfkvy5l643lydnw9re59gtzzwf5mdq'
        ) is None
    # litecoin address
    assert bitcoin.validate_addr('LcNs562P7ZEtAhRcQTabwNEa9QETvEiLzz') is None
    # litecoin P2SH address
    assert bitcoin.validate_addr('MV5rN5EcX1imDS2gEh5jPJXeiW5QN8YrK3') is None



from decimal import Decimal as D

import pytest
from bitcoin.core import COIN
from cashaddress.convert import to_cash_address

from moonwalking.main import BitcoinCash
from moonwalking import wallets
from moonwalking.blocks.exc import NotEnoughAmountError
from moonwalking.wallets import create_addr


def test_create_addr(fee_mocker):
    addr, priv = create_addr('bch')
    assert isinstance(addr, str)
    assert isinstance(priv, str)


@pytest.mark.skip("Broken and we don't know why yet")
async def test_basic_bitcoin_cash_ops(fee_mocker):
    bitcoin_cash = BitcoinCash()
    addr1, priv1 = await wallets.create_wallet('bch')
    addr2, priv2 = await bitcoin_cash.create_wallet()
    addr3, priv3 = await bitcoin_cash.create_wallet()

    assert wallets.validate_addr('bch', addr1)
    assert bitcoin_cash.validate_addr(addr2)
    assert bitcoin_cash.validate_addr(addr3)

    await bitcoin_cash.post('sendtoaddress', addr1, 10000 / COIN)
    await bitcoin_cash.post('generate', 1)
    assert (await bitcoin_cash.get_balance(addr1)) == D(10000) / COIN
    assert (await bitcoin_cash.get_balance(
        to_cash_address(addr1)
    )) == D(10000) / COIN

    tx_id = await bitcoin_cash.send_money(
        priv1,
        [(addr2, D(500) / COIN), (to_cash_address(addr3), D(1500) / COIN)]
    )
    assert tx_id

    await bitcoin_cash.post('generate', 1)
    # 500 fee splited equally onto 2 recievers:
    assert (await bitcoin_cash.get_balance(addr1)) == D(10000 - 2000) / COIN
    assert (await bitcoin_cash.get_balance(addr2)) == D(500 - 250) / COIN
    assert (await bitcoin_cash.get_balance(addr3)) == D(1500 - 250) / COIN

    addr1 = to_cash_address(addr1)
    addr2 = to_cash_address(addr2)
    addr3 = to_cash_address(addr3)

    assert (await bitcoin_cash.get_balance(addr1)) == D(10000 - 2000) / COIN
    assert (await bitcoin_cash.get_balance(addr2)) == D(500 - 250) / COIN
    assert (await bitcoin_cash.get_balance(addr3)) == D(1500 - 250) / COIN


@pytest.mark.skip("Broken and we don't know why yet")
async def test_basic_bitcoin_cash_zero_wallet(fee_mocker):
    bitcoin_cash = BitcoinCash()
    addr1, priv1 = await bitcoin_cash.create_wallet()
    addr2, priv2 = await bitcoin_cash.create_wallet()

    assert bitcoin_cash.validate_addr(addr1)
    assert bitcoin_cash.validate_addr(addr2)

    await bitcoin_cash.post('sendtoaddress', addr1, 10000 / COIN)
    await bitcoin_cash.post('generate', 1)
    assert (await bitcoin_cash.get_balance(addr1)) == D(10000) / COIN

    tx_id = await bitcoin_cash.send_money(priv1, [(addr2, D(10000) / COIN)])
    assert tx_id

    await bitcoin_cash.post('generate', 1)
    # 500 fee splited equally onto 2 recievers:
    assert (await bitcoin_cash.get_balance(addr2)) == D(10000 - 500) / COIN


async def test_new_wallet_empty():
    bitcoin_cash = BitcoinCash()
    addr1, priv1 = await bitcoin_cash.create_wallet()
    assert bitcoin_cash.validate_addr(addr1)
    assert (await bitcoin_cash.get_balance(addr1)) == D(0)


async def test_insufficient_amount():
    bitcoin_cash = BitcoinCash()
    addr1, priv1 = await bitcoin_cash.create_wallet()
    addr2, priv2 = await bitcoin_cash.create_wallet()
    assert bitcoin_cash.validate_addr(addr1)
    await bitcoin_cash.post('sendtoaddress', addr1, 10000 / COIN)
    with pytest.raises(NotEnoughAmountError):
        await bitcoin_cash.send_money(priv1, [(addr2, D(15000) / COIN)])


async def test_insufficient_amount_to_cover_fee():
    bitcoin_cash = BitcoinCash()
    addr1, priv1 = await bitcoin_cash.create_wallet()
    addr2, priv2 = await bitcoin_cash.create_wallet()
    assert bitcoin_cash.validate_addr(addr1)
    await bitcoin_cash.post('sendtoaddress', addr1, 10000 / COIN)
    with pytest.raises(NotEnoughAmountError):
        await bitcoin_cash.send_money(priv1, [(addr2, D(10001) / COIN)])



import os
import json


cd = os.path.dirname
ROOT_DIR = cd(__file__)

LITECOIN_FEE = os.environ.get('LITECOIN_FEE') or '10'  # satoshi / byte
LITECOIN_URL = os.environ.get('LITECOIN_URL')
BITCOIN_CASH_FEE = os.environ.get('BITCOIN_CASH_FEE') or '10'  # satoshi / byte
BITCOIN_CASH_URL = os.environ.get('BITCOIN_CASH_URL')
BITCOIN_URL = os.environ.get('BITCOIN_URL')
BITCOIN_FEE = os.environ.get('BITCOIN_FEE')  # satoshi / byte
BITCOIN_FEE_URL = os.environ.get('BITCOIN_FEE_URL')
ETH_FEE = os.environ.get('ETH_FEE')  # gas price in gwei
ETH_URL = os.environ.get('ETH_URL')
ETH_CHAIN_ID = os.environ.get('ETH_CHAIN_ID') or '4'

BUFFER_ETH_PRIV = os.environ.get('BUFFER_ETH_PRIV')
LND_WALLETS_TOPUP_TRANS_NO = os.environ.get('LND_WALLETS_TOPUP_TRANS_NO', '10')
LND_CONTRACT_ADDR = os.environ.get('LND_CONTRACT_ADDR')
USE_TESTNET = (os.environ.get('USE_TESTNET') or '1') == "1"
FEE_CACHE_TIME = (os.environ.get('FEE_CACHE_TIME') or '10')


COMPILED_CONTRACT_JSON = os.path.join(ROOT_DIR, 'LendingBlockToken.json')

with open(COMPILED_CONTRACT_JSON) as fp:
    LND_CONTRACT = json.load(fp)


assert BITCOIN_FEE is None or BITCOIN_FEE.isdigit(), \
    'BITCOIN_FEE must be None or an integer'
assert ETH_FEE is None or ETH_FEE.isdigit(), \
    'ETH_FEE must be None or an integer'
assert BITCOIN_CASH_FEE.isdigit(), 'BITCOIN_CASH_FEE must be an integer'
assert LITECOIN_FEE.isdigit(), 'LITECOIN_FEE must be an integer'
assert FEE_CACHE_TIME.isdigit(), 'FEE_CACHE_TIME must be an integer'



import logging
from decimal import Decimal as D
from typing import List, Tuple

from aiohttp.client import ClientSession

from eth_abi.abi import decode_abi
from eth_account import Account
from eth_hash.auto import keccak
from eth_utils import from_wei, to_checksum_address, is_address
from eth_utils.currency import to_wei

from hexbytes.main import HexBytes

from .. import settings
from .exc import (
    EthereumError, ReplacementTransactionError, NotEnoughAmountError
)
from .fee import FeeStation
from .base import BaseBlock

logger = logging.getLogger(__name__)
DECIMALS = pow(10, 18)


class EthereumGeneric(BaseBlock):
    MAX_FEE = 100
    URL = settings.ETH_URL
    NETWORK = 'testnet' if settings.USE_TESTNET else 'mainnet'
    MIN_GAS = 21000
    MAX_GAS = 100000

    def get_data(self, method, *params):
        return {
            'jsonrpc': '2.0',
            'method': method,
            'params': list(params),
            'id': self.NETWORK
        }

    async def post(self, *args):
        async with ClientSession() as sess:
            async with sess.post(self.URL, json=self.get_data(*args)) as res:
                resp_dict = await res.json()
                result = resp_dict.get('result')
                error = resp_dict.get('error')
                if error:
                    message = error.get('message')
                    if message == 'replacement transaction underpriced':
                        raise ReplacementTransactionError
                    raise EthereumError(data=resp_dict)
                return result

    async def get_gas_price(self) -> int:
        if not settings.ETH_FEE:
            fee_station = FeeStation('eth')
            transaction_fee = await fee_station.get_fee()
            return min(self.MAX_FEE, transaction_fee)
        return to_wei(int(settings.ETH_FEE), 'gwei')

    async def get_eth_balance(self, addr):
        balance = await self.post('eth_getBalance', addr, 'latest')
        return D(from_wei(int(balance, 16), 'ether'))

    async def get_transaction_dict(self, priv, addr_to, amount, nonce, data,
                                   subtract_fee):
        if data:  # Pull this out.
            gas = self.MAX_GAS
        else:
            get_code = await self.post('eth_getCode', addr_to, 'latest')
            gas = 50000 if len(get_code) > 3 else self.MIN_GAS

        gas_price = await self.get_gas_price()
        amount = to_wei(amount, 'ether')
        if subtract_fee:
            fee = gas * gas_price
            amount -= fee
        return {
            'from': Account.privateKeyToAccount(priv).address,
            'to': addr_to,
            'value': amount,
            'gas': gas,
            'gasPrice': gas_price,
            'data': data,
            'chainId': int(settings.ETH_CHAIN_ID),
            'nonce': nonce
        }

    def make_lnd_transfer_data(self, addr_to, amount):
        # Todo: Generalise to any contract.
        method_hash = self.get_method_hash('transfer')
        addr_hash = self.get_addr_hash(addr_to)
        amount_hash = self.get_amount_hash(amount)
        return method_hash + addr_hash + amount_hash

    async def validate_balance(self, priv, addrs):
        addr_from = Account.privateKeyToAccount(priv).address
        balance = await self.get_eth_balance(addr_from)
        total_amount = sum(amount for addr, amount in addrs)
        if total_amount > balance:
            raise NotEnoughAmountError()

    async def get_transaction_count(self, addr_from):
        nonce = await self.post('eth_getTransactionCount', addr_from,
                                'pending')
        return int(nonce, 16)

    def validate_addr(self, addr):
        if is_address(addr):
            return addr

    def create_addr(self):
        account = Account().create()
        return account.address, account.privateKey.hex()

    @staticmethod
    def get_contract_addr():
        """
        to make tests mocking easier
        """
        return to_checksum_address(settings.LND_CONTRACT_ADDR)

    def get_method_hash(self, method):
        method_sig = self.get_method_signature(method)
        if method_sig:
            return '0x' + keccak(method_sig.encode()).hex()[:8]

    @staticmethod
    def get_addr_hash(addr):
        if addr.startswith('0x'):
            return addr.lower()[2:].zfill(64)
        return ''

    @staticmethod
    def get_amount_hash(num):
        return hex(int(num * DECIMALS))[2:].zfill(64)

    @staticmethod
    def get_method_signature(method_name):
        method_abi = next(x for x in settings.LND_CONTRACT['abi']
                          if x.get('name') == method_name)
        args = ','.join(i['type'] for i in method_abi.get('inputs', ()))
        return f'{method_name}({args})'

    async def call_contract_method(self, method, to_int=False,
                                   to_string=False):
        contract_address = self.get_contract_addr()
        result = await self.post('eth_call', {
            'to': contract_address,
            'data': self.get_method_hash(method)
        }, 'latest')
        if to_int:
            return int(result, 16)
        elif to_string:
            return decode_abi(['string'], HexBytes(result))[0].decode()
        return result

    async def send_eth(self, priv, addrs):
        tx = await EthereumGeneric.build_tx(self, priv, addrs)
        signed = self.sign_tx(priv, tx)
        return await self.broadcast_tx(signed)

    async def send_all_eth_to_buffer_wallet(self, priv):
        addr = Account.privateKeyToAccount(priv).address
        buffer_addr = Account.privateKeyToAccount(
            settings.BUFFER_ETH_PRIV
        ).address
        balance = await self.get_eth_balance(addr)
        return await self.send_eth(priv, [(buffer_addr, balance)])

    async def build_tx(self, priv: str, addrs: List[Tuple[str, D]]):
        await self.validate_balance(priv, addrs)
        addr_from = Account.privateKeyToAccount(priv).address
        nonce = await self.get_transaction_count(addr_from)
        return [
            (await self.get_transaction_dict(
                priv,
                addr,
                amount,
                nonce + i,
                '',
                subtract_fee=True,
                ))
            for i, (addr, amount) in enumerate(addrs)
        ]

    def sign_tx(self, priv, tx):
        return [
            Account.signTransaction(tx_dict, priv).rawTransaction.hex()
            for tx_dict in tx
        ]

    async def broadcast_tx(self, tx):
        # Todo: Retries.
        return [
            (await self.post('eth_sendRawTransaction', tx_hex))
            for tx_hex in tx
        ]



from decimal import Decimal as D

import pytest
from bitcoin.core import COIN

from moonwalking.main import Bitcoin
from moonwalking.blocks.exc import NotEnoughAmountError
from moonwalking.wallets import create_addr


def test_create_addr(fee_mocker):
    addr, priv = create_addr('btc')
    assert isinstance(addr, str)
    assert isinstance(priv, str)


async def test_basic_bitcoin_ops(fee_mocker):
    bitcoin = Bitcoin()
    addr1, priv1 = await bitcoin.create_wallet()
    addr2, priv2 = await bitcoin.create_wallet()
    addr3, priv3 = await bitcoin.create_wallet()

    assert bitcoin.validate_addr(addr1)
    assert bitcoin.validate_addr(addr2)
    assert bitcoin.validate_addr(addr3)

    await bitcoin.post('sendtoaddress', addr1, 20000 / COIN)
    await bitcoin.post('generate', 1)
    assert await bitcoin.get_balance(addr1) == D(20000) / COIN

    tx_id = await bitcoin.send_money(
        priv1,
        [(addr2, D(5000) / COIN), (addr3, D(15000) / COIN)]
    )
    assert tx_id
    await bitcoin.post('generate', 1)

    # 500 fee splited equally onto 2 recievers:
    assert (await bitcoin.get_balance(addr1)) == D(0)
    assert (await bitcoin.get_balance(addr2)) == D(5000 - 250) / COIN
    assert (await bitcoin.get_balance(addr3)) == D(15000 - 250) / COIN


async def test_new_wallet_empty():
    bitcoin = Bitcoin()
    addr1, priv1 = await bitcoin.create_wallet()
    assert bitcoin.validate_addr(addr1)
    assert (await bitcoin.get_balance(addr1)) == D(0)


async def test_insufficient_amount(fee_mocker):
    bitcoin = Bitcoin()
    addr1, priv1 = await bitcoin.create_wallet()
    addr2, priv2 = await bitcoin.create_wallet()
    assert bitcoin.validate_addr(addr1)
    await bitcoin.post('sendtoaddress', addr1, 10000 / COIN)
    with pytest.raises(NotEnoughAmountError):
        await bitcoin.send_money(priv1, [(addr2, D(15000) / COIN)])


async def test_insufficient_amount_to_cover_fee(fee_mocker):
    bitcoin = Bitcoin()
    addr1, priv1 = await bitcoin.create_wallet()
    addr2, priv2 = await bitcoin.create_wallet()
    assert bitcoin.validate_addr(addr1)
    await bitcoin.post('sendtoaddress', addr1, 10000 / COIN)
    with pytest.raises(NotEnoughAmountError):
        await bitcoin.send_money(priv1, [(addr2, D(10001) / COIN)])
